from aiger_bv import atom

from aiger_sat import sat_bv


def test_get_model():
    expr = atom(4, 'x') & atom(4, 'y') < 2
    f = sat_bv.SolverBVWrapper()
    f.add_expr(expr)

    model = f.get_model()
    assert len(model) == 2
    assert expr(model)


def test_solve():
    expr = atom(4, 'x') & atom(4, 'y') < 2
    model = sat_bv.solve(expr)
    assert len(model) == 2
    assert expr(model)



import aiger

import aiger_sat
from aiger_sat import SolverWrapper


def test_smoke():
    x, y, z = map(aiger.atom, ['x', 'y', 'z'])
    solver = SolverWrapper()

    expr = (x & y) | ~z
    solver.add_expr(expr)
    assert solver.inputs == {'x', 'y', 'z'}
    assert len(solver.sym_table) == 3

    assert all(v in solver.sym_table for v in 'xyz')

    assert solver.unsolved

    assert solver.is_sat()

    assert not solver.unsolved
    model = solver.get_model()
    assert model is not None
    assert len(model) == 3
    assert expr(model)

    assert not solver.is_sat(assumptions={
        'x': False,
        'z': True,
    })
    assert not solver.unsolved

    core = solver.get_unsat_core()
    assert core == {'x': False, 'z': True}


def test_empty_unsat_core():
    x = aiger.atom('x')

    solver = SolverWrapper()
    solver.add_expr(x & ~x)

    assert solver.get_model() is None
    assert solver.get_unsat_core() == {}


def test_solve():
    x, y, z = map(aiger.atom, ['x', 'y', 'z'])
    model = aiger_sat.solve(x & y & z)
    assert model == {'x': True, 'y': True, 'z': True}


def test_is_sat():
    x, y, z = map(aiger.atom, ['x', 'y', 'z'])
    assert aiger_sat.is_sat(x & y & z)


def test_is_valid():
    assert aiger_sat.is_valid(aiger.atom(True))


def test_equiv():
    x, y, z = map(aiger.atom, ['x', 'y', 'z'])
    assert aiger_sat.are_equiv(x & y, x & y & (z | ~z))



try:
    import aiger_bv
except ModuleNotFoundError:
    raise ModuleNotFoundError("Please install with bitvector option.")

import attr
from pysat.solvers import Glucose4

import aiger_sat


@attr.s(slots=True, auto_attribs=True)
class SolverBVWrapper(aiger_sat.SolverWrapper):
    bmap: aiger_bv.bundle.BundleMap = aiger_bv.bundle.BundleMap()

    def add_expr(self, expr):
        super().add_expr(expr)
        self.bmap += expr.aigbv.imap

    def get_model(self):
        model = {n: size*(False,) for n, size in self.bmap.items()}
        model = self.bmap.blast(model)
        model.update(super().get_model())
        return self.bmap.unblast(model)


def solve(expr, *, engine=Glucose4):
    return aiger_sat.solve(expr, engine=engine, wrapper=SolverBVWrapper)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiger_sat']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=20.0.0,<21.0.0',
 'py-aiger-cnf>=5.0.0,<6.0.0',
 'py-aiger>=6.0.0,<7.0.0',
 'python-sat>=0.1.5,<0.2.0']

extras_require = \
{'bitvector': ['py-aiger-bv>=4.0.0,<5.0.0']}

setup_kwargs = {
    'name': 'py-aiger-sat',
    'version': '3.0.2',
    'description': 'Pythonic interface between AIGs and SAT solvers.',
    'long_description': "# py-aiger-sat\nPythonic interface between AIGs and SAT solvers.\n\n[![Build Status](https://cloud.drone.io/api/badges/mvcisback/py-aiger-sat/status.svg)](https://cloud.drone.io/mvcisback/py-aiger-sat)\n[![codecov](https://codecov.io/gh/mvcisback/py-aiger-sat/branch/master/graph/badge.svg)](https://codecov.io/gh/mvcisback/py-aiger-sat)\n[![PyPI version](https://badge.fury.io/py/py-aiger-sat.svg)](https://badge.fury.io/py/py-aiger-sat)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\n<!-- markdown-toc start - Don't edit this section. Run M-x markdown-toc-generate-toc again -->\n**Table of Contents**\n\n- [Installation](#installation)\n- [Usage](#usage)\n\n<!-- markdown-toc end -->\n\n\n# Installation\n\nIf you just need to use `aiger_sat`, you can just run:\n\n`$ pip install py-aiger-sat`\n\nFor developers, note that this project uses the\n[poetry](https://poetry.eustace.io/) python package/dependency\nmanagement tool. Please familarize yourself with it and then\nrun:\n\n`$ poetry install`\n\n# Usage\n\n`aiger_sat` has two seperate API's. The first, called the Object API,\ncenters around the `SolverWrapper` object - a thin wrapper around a\n`pysat` solver. The second is a Function API which exposes 4 functions\n`solve`, `is_sat`, `is_valid`, and `are_equiv`. The function API is\nprimarily useful for simple 1-off SAT instances, where as the object\nAPI is more useful when incremental solves are needed, or the\nunderlying `pysat` solver must be exposed.\n\n## Object API\n\n```python\nfrom aiger_sat import SolverWrapper\n\nsolver = SolverWrapper()  # defaults to Glucose4\n\nfrom pysat.solvers import Glucose3\nsolver2 = SolverWrapper(solver=Glucose3)\n```\n\n`solver` operate on boolean expressions in the form of `aiger`\ncircuits with a single output. For example,\n\n\n```python\nimport aiger\n\nx, y, z = map(aiger.atom, ['x', 'y', 'z'])\n\nexpr = (x & y) | ~z\nsolver.add_expr(expr)\nassert solver.is_sat()\nmodel = solver.get_model()\nprint(model)  # {'x': True, 'y': False, 'z': False}\nassert expr(model)\n```\n\nFurther, `aiger_sat` supports making assumptions and computing\nunsat_cores.\n\n```python\n# Make invalid assumption.\nassert not solver.is_sat(assumptions={\n    'x': False,\n    'z': True,\n})\nassert not solver.unsolved\n\ncore = solver.get_unsat_core()\nassert core == {'x': False, 'z': True}\n```\n\n## Function API\n\n```python\nimport aiger\nimport aiger_sat\n\nx, y, z = map(aiger.atom, ['x', 'y', 'z'])\nassert aiger_sat.is_sat(x & y & z)\n\nmodel = aiger_sat.solve(x & y & z)\nassert model == {'x': True, 'y': True, 'z': True}\n\nassert aiger_sat.is_valid(aiger.atom(True))\n\nexpr1 = x & y\nexpr2 = x & y & (z | ~z)\nassert aiger_sat.are_equiv(expr1, expr2)\n```\n\n## BitVector Support\n\n`py-aiger-sat` also natively supports the `py-aiger-bv` bitvector\nlibrary.\n\nTo enable this support, make sure that `py-aiger-bv` is installed,\neither manually:\n\n`$ pip install py-aiger-bv`\n\nor by installing `py-aiger-sat` with the `bitvector` option:\n\n`$ pip install py-aiger-sat[bitvector]` or `$ poetry install --extras=bitvector`\n\nUsage is analogous to the non-bitvector usage.\n\n```python\nfrom aiger_bv import atom\nfrom aiger_sat import sat_bv\n\n# Object API\nexpr = atom(4, 'x') & atom(4, 'y') < 2\nf = sat_bv.SolverBVWrapper()\nf.add_expr(expr)\n\nmodel = f.get_model()\n\n# Function API.\nmodel = sat_bv.solve(expr)\n\nprint(model)\n# {'x': (False, False, True, True), 'y': (False, False, True, True)}\n```\n",
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'mvc@linux.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mvcisback/py-aiger-sat',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from functools import wraps
from typing import Callable, Hashable, TypeVar, Set

import aiger
import attr
import funcy as fn
from aiger_cnf import aig2cnf
from pysat.solvers import Glucose4
from bidict import bidict


Solver = TypeVar("Solver")


def _require_solved(func):
    @wraps(func)
    def decorated(self, *args, **kwargs):
        if self.unsolved:
            self.solver.solve()

        return func(self, *args, **kwargs)
    return decorated


def id_pool_factory():
    max_var = 0

    @fn.memoize
    def _fresh(val):
        nonlocal max_var
        max_var += 1
        return max_var

    return _fresh


@attr.s(auto_attribs=True)
class SolverWrapper:
    solver: Solver = attr.ib(factory=Glucose4)
    unsolved: bool = True
    inputs: Set[str] = attr.ib(factory=set)
    sym_table: bidict = attr.ib(factory=bidict)
    _id_pool: Callable[[Hashable], int] = attr.ib(factory=id_pool_factory)

    def add_expr(self, expr):
        expr = aiger.BoolExpr(expr.aig)
        self.unsolved = True
        self.inputs |= expr.inputs
        cnf = aig2cnf(expr, fresh=self._id_pool)
        for clause in cnf.clauses:
            self.solver.add_clause(clause)
        self.sym_table.update(cnf.input2lit)

    def is_sat(self, assumptions=None):
        if assumptions is None:
            assumptions = {}

        assumptions = [
            (1 if v else -1)*self.sym_table[k] for k, v in assumptions.items()
        ]

        self.unsolved = False
        return self.solver.solve(assumptions)

    def _translate(self, cube):
        if cube is None:
            return cube

        idx2sym = self.sym_table.inv
        return {
            idx2sym[abs(v)]: v > 0 for v in cube
            if abs(v) in idx2sym and idx2sym[abs(v)] in self.inputs
        }

    @_require_solved
    def get_model(self):
        model = self.solver.get_model()
        return self._translate(model)

    @_require_solved
    def get_unsat_core(self):
        core = self._translate(self.solver.get_core())
        return {} if (core is None) else core


def _solve(expr, method, engine, wrapper=SolverWrapper):
    solver = wrapper(engine())
    solver.add_expr(expr)
    return method(solver)


def solve(expr, *, engine=Glucose4, wrapper=SolverWrapper):
    return _solve(expr, wrapper.get_model, engine=engine, wrapper=wrapper)


def is_sat(expr, *, engine=Glucose4):
    return _solve(expr, SolverWrapper.is_sat, engine=engine)


def is_valid(expr, *, engine=Glucose4):
    return not is_sat(~expr, engine=engine)


def are_equiv(expr1, expr2, *, engine=Glucose4):
    # Make sure they're expressions.
    assert len(expr1.aig.outputs) == len(expr2.aig.outputs) == 1
    expr1, expr2 = map(lambda e: aiger.BoolExpr(e.aig), (expr1, expr2))

    return is_valid(expr1 == expr2)
